function record_perimetry
% Visual fields testing program, written by Matt J Dunn and Nikita Thomas

%% Parameter setup
calibrateEyeTracker = true;
manualFixationOffset_deg = [0,0]; % adjust here for a preferred retinal locus
wavesToRun = {'first','second'}; % which waves should be run?

stimulusDuration_ms = 200;
stimulusSize_deg = 0.43;

CentralFixationTarget.on = true; % set to true to have a fixation target in the centre
CentralFixationTarget.type = 'standard';
CentralFixationTarget.colour = [0.5,0,0]; % colour of central fixation target in RGB
CentralFixationTarget.size_deg = 0.25; % size of the central fixation target

FixationReminderAnnulus.tolerance_deg = 3; % fixation reminder will appear if the gaze deviates further than this from screen centre
FixationReminderAnnulus.thickness_deg = 0.2;
FixationReminderAnnulus.colour_rgb = [0.2,0,0]; % annulus will be red

numberofStimulusPresentationsbetweenDriftCorrections = 100;

QuestParameters.estimatedSD = 0.2;
QuestParameters.pThreshold = 0.5; % probability of positive response (50% for most perimetry programs)
QuestParameters.beta = 3.5; % steepness of psychometric function (default for QUEST = 3.5)
QuestParameters.delta = 0.01; % fraction of trials in which observer presses blindly (default for QUEST = 0.01)
QuestParameters.gamma = 0; % probability of success at zero intensity (2AFC = 0.5, yes/no = 0)
questTerminationMaximumIntervalSize = 0.2; % QUEST can terminate when the interval is <= 0.2 which is when the algorithm is 90% confident that threshold is within 1cd/m2 of the true threshold

StaircaseParameters.useStaircasesIfQuestFails = true; % can turn off staircasing completely here. Otherwise, if QUEST has been too hasty in determining threshold and there hasn't been any reversal of the stimulus intensity direction, then we might switch to a staircase. This is one of the main reasons why we count "StimulusList_main(stimulusIndex).nUp" and "StimulusList_main(stimulusIndex).nDown" in "wave.m" (lines 140-145 in "wave.m"). If there has been no reversal of stimulus intensity and QUEST has already determined threshold then the stimulus should be switched to a staircase. This is because QUEST has shown to be occasionally too fast in its calculation of threshold and having this as an extra step safegards against this potentially happening (particularly when we are using a different method of threshold termination and not terminating after a fixed number of presentations).
nReversalsRequired = 2;
maximumLuminance_cdm2 = 85; % this is the reference point required to calculate logarithmic steps in a staircase.
backgroundLuminance_cdm2 = 1.27;

nFalsePositiveTrials = 5;
nFalseNegativeTrials = 5;

startingIntensityAboveNormativeData_logarithmic = 15; % how far above the normative (expected) thresholds we make the first presentation at each stimulus location.
falsePOSfalseNEG_logIncrement = 9; % this is the log increment required for false positive and false negative catch trials

additionToInterstimulusInterval_sec = 0.7; % a fixed amount of time top be added to the mean participant response time when calculating dynamic intervals. The concept of using dynamic intervals in perimetry was first described by Boer et al. 1982 where they increased whatever the interstimulus interval was by 0.7s. This is because even a 0.5s or 1s interstimulus interval is still quite fast so every calculation of the final interstimulus interval is augmented by 0.7s
dynamicIntervalRange_min_max_sec = [0.35,2]; % minimum and maximum possible interstimulus interval in seconds, after taking into account both participant response times so far, and the fixed addition above
falsePositiveCatchTime_sec = 0.18; % any responses made within this time window are considered 'trigger happy' and marked as a false positive

addpath(genpath('required_files')); % add resources folder

ScreenInfo = oms_getscreeninfo;

backgroundLuminance_AU = 1 - (ScreenInfo.maxLuminance_cdm2 - backgroundLuminance_cdm2) / (ScreenInfo.maxLuminance_cdm2 - ScreenInfo.minLuminance_cdm2); % calculate background luminance as a fraction of monitor output
if isnan(backgroundLuminance_AU)
    warning('Unable to set background luminance; please specify screen luminance range in your screens.csv file')
    backgroundLuminance_AU = 0; % default to minimum monitor luminance
end

% Set default calibration parameters (replace here if using a custom calibration, e.g. for nystagmus)
calibrationMatrix = [];
calibrationType = 'tracker_default';

gazeContingencyType = 'Continuous gaze-contingency (stimulus moving)'; % can also select 'No gaze-contingency'
eye = questdlg('Which eye?','Eye?','Right','Left','Right');
eye = lower(eye);
gridType = '10-2';

%% Launch program
participantId = inputdlg('Please enter participant ID','Participant ID');
participantId = participantId{:}; % extract contents

patientAge = []; % init
while isempty(patientAge) || isnan(patientAge) % TODO: I've temporarily turned off the age request as it's not needed until we restore age-matched norms
    patientAge = inputdlg('Please enter patient age','Age');
    patientAge = str2double(patientAge{:}); % extract contents
end

% Setup output file name and directory
finalFileName = [participantId,'_perimetryEyeTrace']; % the name of the output file
finalFileDirectory = fullfile('Data',participantId); % work out the directory we will use for the data
finalFile = fullfile(finalFileDirectory,finalFileName); % work out the complete filename
[~,~] = mkdir(finalFileDirectory); % create directory

% Load screen and stimulus properties
[tempEyeTraceFileName,tracker,ScreenInfo,window] = oms_eyetracker_startup([],calibrateEyeTracker,[],[],ScreenInfo.name,backgroundLuminance_AU,[],true,eye); % launch eyetracker, and do not shut down screen afterwards

% Setup FixationReminderAnnulus
FixationReminderAnnulus.diameter_px = oms_convert_degrees2px(FixationReminderAnnulus.tolerance_deg * 2,ScreenInfo); % convert to px (double the size, as the annulus is a full circle, and we have only specified radius)
FixationReminderAnnulus.thickness_px = oms_convert_degrees2px(FixationReminderAnnulus.thickness_deg,ScreenInfo); % convert to px
FixationReminderAnnulus.rect = [0,0,FixationReminderAnnulus.diameter_px,FixationReminderAnnulus.diameter_px]; % define a rect for circle (currently in top-left of screen)
FixationReminderAnnulus.rect = CenterRectOnPointd(FixationReminderAnnulus.rect,ScreenInfo.width_px/2,ScreenInfo.height_px/2); % move the rect to the screen center

%% Prepare main stimulus list
stimulusGridFileContents = csvread(which(['grid_',gridType,'.csv']),1); % read grid CSV file (ignore first row)

nStimulusPositions_main = size(stimulusGridFileContents(:,1),1); % count number of rows

StimulusList_main(1:nStimulusPositions_main) = struct('position_deg',[],'rect',nan(1,4),'QuestObject',[],'presentAsStaircase',false,'falsePositivePresentationDue',false,'falseNegativePresentationDue',false,'nUp',0,'nDown',0,'StaircaseObject',[],'nPresentations',0,'Presentations',struct('timestamp',[],'stimulusIntensity',[],'trialType',[],'seen',[],'reactionTime_ms',[])); % init struct that will hold details about every stimulus position

normativeThresholdCoefficients = stimulusGridFileContents(:,3:4);
for i = 1:nStimulusPositions_main
    StimulusList_main(i).position_deg = stimulusGridFileContents(i,1:2);
    StimulusList_main(i).rect = calculatestimulusrect(StimulusList_main(i).position_deg,stimulusSize_deg,ScreenInfo);
    predictedThreshold_maiaDecibels = normativeThresholdCoefficients(i,1) * patientAge + normativeThresholdCoefficients(i,2); % get normative value (expected threshold) in cd/m^2 (multiplication variables * patientAge + add, i.e. y = mx+c)
    predictedThreshold_cdm2 = log10(313.8/predictedThreshold_maiaDecibels); % convert from MAIA dB into cdm2 % TODO: This extraneous step could be removed by converting the equations in the grid CSVs into cdm2
    if predictedThreshold_cdm2 <= backgroundLuminance_cdm2
        warning('Predicted thresholds are too low! Coercing to background luminance...')
        predictedThreshold_cdm2 = backgroundLuminance_cdm2;
    elseif isinf(predictedThreshold_cdm2)
        warning('Predicted thresholds are infinte! Does your grid file contain normative value equations? Arbitrarily setting predicted threshold to 5cdm2...')
        predictedThreshold_cdm2 = 5;
    end
    predictedThreshold_au = convert_cdm22au(predictedThreshold_cdm2,ScreenInfo); % convert from cd/m^2 to computer screen output luminance values (0-1)
    StimulusList_main(i).QuestObject = QuestCreate(predictedThreshold_au,QuestParameters.estimatedSD,QuestParameters.pThreshold,QuestParameters.beta,QuestParameters.delta,QuestParameters.gamma);
    StimulusList_main(i).QuestObject.normalizePdf = true; % this adds a few ms per call to QuestUpdate, but otherwise the pdf will underflow after about 1000 trials.
end

%% Prepare catch trials
nCatchTrials = nFalsePositiveTrials + nFalseNegativeTrials;
catchTrialIndices = randsample(1:nStimulusPositions_main,nCatchTrials); % pick trial numbers at random to be catch trials
falsePositiveTrialIndices = catchTrialIndices(1:nFalsePositiveTrials); % take some as false positives
falseNegativeTrialIndices = catchTrialIndices(nFalsePositiveTrials + 1:end); % ...and the rest as false negatives

[StimulusList_main(falsePositiveTrialIndices).falsePositivePresentationDue] = deal(true); % set a flag on these trials, so that a false positive is performed during the second wave
[StimulusList_main(falseNegativeTrialIndices).falseNegativePresentationDue] = deal(true); % set a flag on these trials, so that a false negative is performed during the second wave

%% Initialise variables
nFalsePositiveResponses = 0; % init
nFalseNegativeResponses = 0; % init
driftCorrectionOffset_deg = [0,0]; % init
startTimestamp = GetSecs; % grab the initial timestamp so we can calculate time within the test at each subsequent call of GetSecs()
reactionTimeRecord_sec = []; % a vector containing all reaction times in sec. Used to store the last few reaction times, regardless of where the stimulus was

escapePressed = false; % init

%% Two waves
for waveName = wavesToRun
    if ~escapePressed
        fprintf('\nBeginning %s stimulus wave...',waveName{:})
        [StimulusList_main,reactionTimeRecord_sec,escapePressed,tempEyeTraceFileName,window] = wave(waveName{:},numberofStimulusPresentationsbetweenDriftCorrections,tracker,window,ScreenInfo,eye,startTimestamp,stimulusDuration_ms,CentralFixationTarget,backgroundLuminance_AU,gazeContingencyType,FixationReminderAnnulus,calibrationMatrix,driftCorrectionOffset_deg,manualFixationOffset_deg,StimulusList_main,StaircaseParameters,reactionTimeRecord_sec,nFalsePositiveResponses,nFalseNegativeResponses,falsePositiveCatchTime_sec,dynamicIntervalRange_min_max_sec,additionToInterstimulusInterval_sec,questTerminationMaximumIntervalSize,calibrationType,tempEyeTraceFileName,startingIntensityAboveNormativeData_logarithmic,nReversalsRequired,maximumLuminance_cdm2,falsePOSfalseNEG_logIncrement);
    end
end

%% Shutdown
fprintf('\nShutting down...')
oms_screen_shutdown(ScreenInfo,window,true); % quit with wait screen

fileName_AllData = [finalFile,'_allData.mat'];
save(fileName_AllData,'StimulusList_main');

%% Output presentation/response report
report = []; % init
for StimulusList = {StimulusList_main}
    StimulusList = StimulusList{:}; % extract from cell
    for i = 1:length(StimulusList) % for each presentation...
        clear subreport
        if ~isempty(StimulusList(i).Presentations(1).timestamp) % as long as there was at least one response to a presentation at this location...
            subreport(:,1) = num2cell([StimulusList(i).Presentations.timestamp]);
            subreport(:,2) = num2cell(StimulusList(i).position_deg(1));
            subreport(:,3) = num2cell(StimulusList(i).position_deg(2));
            subreport(:,4) = {StimulusList(i).Presentations.trialType};
            subreport(:,5) = num2cell([StimulusList(i).Presentations.stimulusIntensity]);
            subreport(:,6) = num2cell([StimulusList(i).Presentations.reactionTime_ms]);
            subreport(:,7) = num2cell([StimulusList(i).Presentations.seen]);
            report = [report;subreport]; % append to main report
        end
    end
end
report = sortrows(report,1); % order the report by the timestamp
fileName = [finalFile,'_responses.csv'];
header = 'timestamp_sec,position_x_deg,position_y_deg,trialType,stimulusIntensity,reactionTime_ms,seen';
oms_generatecsv(report,fileName,header);

%% Output threshold report (can be used to generate actual field plots)
clear report
for i = 1:length(StimulusList_main) % for each presentation...
    if ~isempty(StimulusList_main(i).Presentations(1).timestamp) % as long as there were presentations at this location...
        report(i,1) = StimulusList_main(i).position_deg(1);
        report(i,2) = StimulusList_main(i).position_deg(2);
        report(i,3) = QuestMean(StimulusList_main(i).QuestObject);
        report(i,4) = convert_au2cdm2(report(i,3),ScreenInfo); % convert to cd/m^2
    end
end
if exist('report','var') % if the program actually got far enough to give us anything we can use to generate threshold values...
    fileName = [finalFile,'_thresholds.csv'];
    header = 'position_x_deg,position_y_deg,threshold_monitorluminance,threshold_cdm2';
    oms_generatecsv(report,fileName,header);
end

oms_eyetracker_shutdown(tempEyeTraceFileName,finalFile,ScreenInfo);



function rect = calculatestimulusrect(actualStimulusPosition_deg,stimulusSize_deg,ScreenInfo)
% Calculate stimulus rect, correcting for curvature of field

xLeft_deg = actualStimulusPosition_deg(1) - stimulusSize_deg/2; % work out position of rect in degrees
yTop_deg = actualStimulusPosition_deg(2) + stimulusSize_deg/2;
xRight_deg = actualStimulusPosition_deg(1) + stimulusSize_deg/2;
yBottom_deg = actualStimulusPosition_deg(2) - stimulusSize_deg/2;
% NOTE: These aren't the actual POSITIONS that the stimuli are shown in.
% But, they are the actual SIZES of the stimuli. It's important
% here that we know WHERE the stimulus will be shown, only because we need
% to account for curvature of field. But when the stimulus is shown
% on-screen, we just take this rect and center it on the appropriate
% position on-screen (to take account of eye movements). Ideally, we'd
% recalculate the rect at the time of presentation (since if the image is
% shifted on-screen, we should adjust the rect size), but constant
% recalculation could slow program execution too much during high-speed
% gaze contingency.

xLeft_px = oms_convert_degrees2px(xLeft_deg,ScreenInfo) + ScreenInfo.width_px/2; % convert stimulusRect position to degrees
yTop_px = -oms_convert_degrees2px(yTop_deg,ScreenInfo) + ScreenInfo.height_px/2;
xRight_px = oms_convert_degrees2px(xRight_deg,ScreenInfo) + ScreenInfo.width_px/2;
yBottom_px = -oms_convert_degrees2px(yBottom_deg,ScreenInfo) + ScreenInfo.height_px/2;

rect = [xLeft_px,yTop_px,xRight_px,yBottom_px];
