function [ScreenInfo,window,window_overlay] = oms_screen_startup(screenName,bitsSharpMode,backgroundColour_rgb,includeWaitScreen,antialiasingMultisampling,ForcedResolution)
% Oculomotor Suite: setup screen in PTB
% Created by Matt J Dunn

addpath(genpath('required_files')); % must have the 'required_files' subfolder present

if ~exist('screenName','var') || isempty(screenName)
    screenName = [];
end
if ~exist('includeWaitScreen','var') || isempty(includeWaitScreen)
    includeWaitScreen = false;
end
if ~exist('antialiasingMultisampling','var') || isempty(antialiasingMultisampling)
    antialiasingMultisampling = 0; % higher numbers prevent aliasing but consume more video memory
end

ScreenInfo = oms_getscreeninfo(screenName); % grab screen info for this computer

Screen('Preference','SkipSyncTests',0); % do not skip sync tests
    
Screen('Preference','VisualDebugLevel',3); % start up with black screen

PsychDefaultSetup(2); % start Psychtoolbox

fprintf('\nAssuming screen width is %i mm at a distance of %i mm',ScreenInfo.width_mm,ScreenInfo.distance_mm)

ListenChar(2); % disable key output to MATLAB window

if isnan(ScreenInfo.portId) % if a screen number hasn't been pre-specified
    ScreenInfo.portId = max(Screen('Screens')); % ...just set to the primary connected monitor
end

ScreenInfo.oldResolution = Screen('Resolution',ScreenInfo.portId); % store the current resolution as oldResolution (OMS is agnostic as to whether this is actually any different)

[ThisScreenInfo.width_px,ThisScreenInfo.height_px] = Screen('WindowSize',ScreenInfo.portId); % retrieve current screen resolution

if ThisScreenInfo.width_px ~= ScreenInfo.width_px || ThisScreenInfo.height_px ~= ScreenInfo.height_px
    warning('Screen info does not match stored defaults! Please check you have them set correctly in ''/required_files/preferences/screens.csv''')
end

ScreenInfo.width_px = ThisScreenInfo.width_px;
ScreenInfo.height_px = ThisScreenInfo.height_px;
ScreenInfo.rect = [0,0,ThisScreenInfo.width_px,ThisScreenInfo.height_px]; % this one is for convenience later on
ScreenInfo.grey = 0.5; % colour index of grey (in some older code, 128/255 is used)

if exist('bitsSharpMode','var') && ~isempty(bitsSharpMode) % if we have specified a bitsSharpMode
    ScreenInfo.bitsSharpMode = bitsSharpMode; % store in the ScreenInfo for easy access later (in case we need to reboot the screen)
end

if ~exist('backgroundColour_rgb','var') || isempty(backgroundColour_rgb)
    backgroundColour_rgb = ScreenInfo.grey; % if a window colour isn't specified, then use grey
end

PsychImaging('PrepareConfiguration');

if ~isempty(ScreenInfo.bitsSharpMode) && ~strcmp(ScreenInfo.bitsSharpMode,'none') % TODO: Rename the field 'bitsSharpMode' to 'bitDepthMode' as we now allow for either Bits# or VPixx devices; the mode names should be renamed to be generic as well
    try
        VPixx_success = Datapixx('Open'); % first, try to start VPixx device
    catch
        VPixx_success = false;
    end
    if VPixx_success
        highBitDepthDisplayTech = 'vpixx';
    else % if VPixx didn't connect, try Bits#
        try
            CRS_success = BitsPlusPlus('OpenBits#'); % try to start CRS Bits# box
        catch
            CRS_success = false;
        end
        if CRS_success
            highBitDepthDisplayTech = 'bitsSharp';
        else
            highBitDepthDisplayTech = 'none';
        end
    end
    
    if ~strcmp(highBitDepthDisplayTech,'none')
        PsychImaging('AddTask','General','FloatingPoint32Bit'); % use 32-bit precision
    else
        warning('Unable to enter high bit depth mode!')
        ScreenInfo.bitsSharpMode = 'none';
    end
    
    switch highBitDepthDisplayTech
        case 'vpixx'
            switch ScreenInfo.bitsSharpMode
                case 'monoPlusPlus'
                    PsychImaging('AddTask','General','EnableDataPixxM16Output');
                case 'monoPlusPlusWithOverlay'
                    PsychImaging('AddTask','General','EnableDataPixxM16OutputWithOverlay');
                case 'colourPlusPlus'
                    % TODO
                case 'bitsPlusPlus'
                    % TODO
            end
        case 'bitsSharp'
            switch ScreenInfo.bitsSharpMode
                case 'monoPlusPlus'
                    PsychImaging('AddTask','General','EnableBits++Mono++Output');
                case 'monoPlusPlusWithOverlay'
                    PsychImaging('AddTask','General','EnableBits++Mono++OutputWithOverlay');
                case 'bitsPlusPlus'
                    PsychImaging('AddTask','General','EnableBits++Bits++Output');
            end
    end    
end

if ScreenInfo.mirror
    PsychImaging('AddTask','AllViews','FlipHorizontal'); % flip the screen
    fprintf('\nFlipping monitor output horizontally')
end

ScreenInfo.backgroundColour_rgb = backgroundColour_rgb;

window = PsychImaging('OpenWindow',ScreenInfo.portId,ScreenInfo.backgroundColour_rgb,[],[],[],ScreenInfo.stereoMode,antialiasingMultisampling); % 'window' will appear on screen in windowColour
ScreenInfo.interFrameInterval_s = Screen('GetFlipInterval',window);  % query duration of one monitor refresh interval
ScreenInfo.hz = Screen('NominalFrameRate',ScreenInfo.portId); % get refresh rate
if ScreenInfo.hz == 0 % on non-Linux systems, this will be reported as zero, so instead, use 1/interFrameInterval_s
    ScreenInfo.hz = 1/ScreenInfo.interFrameInterval_s;
end

Priority(MaxPriority(window)); % set priority to max the system can handle. If using a slow system, comment this line
Screen('BlendFunction',window,GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA); % enable transparency channels
HideCursor;

if ~isempty(ScreenInfo.bitsSharpMode) && ~strcmp(ScreenInfo.bitsSharpMode,'none') % if we're using a Bits#...
    linearLUT = linspace(0,1,256)' * [1,1,1]; % if using Bits#, we should linearise the computer graphics card gamma table and use Bits#'s own LUT adjustments
    Screen('LoadNormalizedGammaTable',window,linearLUT);
elseif ~isempty(ScreenInfo.gammaTableName) % if a gamma table is specified for this screen (can be used in the absence of a Bits#-type device)
    try
        table2array(readtable(ScreenInfo.gammaTableName)); % read the TXT file and convert to an array
        Screen('LoadNormalizedGammaTable',window,gammaTable * [1 1 1]); % load gamma table and apply to anything shown on screen
    catch
        warning('Unable to load gamma table ''%s''',ScreenInfo.gammaTableName)
    end
end

if strcmp(ScreenInfo.bitsSharpMode,'monoPlusPlusWithOverlay') % if we are using a colour overlay, we need to define a colour lookup table (CLUT) to determine which colours we want to have available for use
    CLUT_overlay = hsv(256); % first load a standard CLUT
    CLUT_overlay(1,:) = ScreenInfo.grey; % set first entry (which is called as entry 0!) to average grey
    CLUT_overlay(2,:) = [1,0,0]; % entry #1 = pure red
    CLUT_overlay(3,:) = [0,1,0]; % entry #2 = pure green
    CLUT_overlay(4,:) = [0,0,1]; % entry #3 = pure blue
    % Note: If you want to define a custom CLUT for an experiment you would need to inject code here to achieve that. I have not yet had need for more than a few basic colours in an overlay which is why this is so simple.
    % NOTE: It is also possible to define a separate CLUT for the console screen, different to the CLUT for the stimulus screen. This allows things to be shown on the console monitor which are transparent on the stimulus monitor, e.g. to allow messages to the experimenter that cannot be seen by the participant. Tutorial at https://www.vpixx.com/manuals/psychtoolbox/html/PROPixxDemo7.html
    switch highBitDepthDisplayTech
        case 'vpixx'
            Datapixx('SetVideoClut',CLUT_overlay);
            Datapixx('RegWr'); % must call this command after any direct calls to the Datapixx API to ensure the register is updated
        case 'bitsSharp'
            Screen('LoadNormalizedGammaTable',window,CLUT_overlay,2);
    end
    window_overlay = PsychImaging('GetOverlayWindow',window); % get handle to the overlay window so we can draw to it
else
    window_overlay = Screen('OpenOffscreenWindow',-1); % if no overlay, create a dummy offscreen window
end

KbName('UnifyKeyNames');

if includeWaitScreen
    textColour_rgb = 1 - backgroundColour_rgb;
    if all(textColour_rgb == ScreenInfo.grey) % if the screen is mid-grey
        textColour_rgb = 0; % make text black
    end
    DrawFormattedText(window,'Press any key to begin','center','center',textColour_rgb);
    Screen('Flip',window);
    oms_awaitkeypress;
    KbReleaseWait;
end